
@file:Suppress("unused")

package ru.dbotthepony.mc.otm.util

import com.github.benmanes.caffeine.cache.Cache
import com.github.benmanes.caffeine.cache.Caffeine
import com.github.benmanes.caffeine.cache.Scheduler
import com.google.common.collect.ImmutableList
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import net.minecraft.Util
import net.minecraft.core.BlockPos
import net.minecraft.core.Holder
import net.minecraft.core.HolderLookup
import net.minecraft.core.Registry
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.network.chat.ComponentContents
import net.minecraft.network.chat.contents.TranslatableContents
import net.minecraft.resources.ResourceKey
import net.minecraft.resources.ResourceLocation
import net.minecraft.tags.TagKey
import net.minecraft.world.entity.Entity
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.component.ItemAttributeModifiers
import net.minecraft.world.item.crafting.CraftingInput
import net.minecraft.world.item.crafting.RecipeInput
import net.minecraft.world.level.BlockGetter
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.level.block.state.StateHolder
import net.minecraft.world.level.block.state.properties.Property
import net.minecraft.world.phys.Vec3
import net.neoforged.neoforge.fluids.FluidStack
import net.neoforged.neoforge.items.IItemHandler
import ru.dbotthepony.mc.otm.util.math.BlockRotation
import ru.dbotthepony.mc.otm.util.math.BlockRotationFreedom
import ru.dbotthepony.mc.otm.util.math.Vector
import java.lang.ref.Reference
import java.math.BigInteger
import java.time.Duration
import java.util.*
import java.util.concurrent.Callable
import java.util.concurrent.Future
import java.util.function.Supplier
import java.util.stream.Stream
import java.util.stream.StreamSupport
import kotlin.jvm.optionals.getOrNull
import kotlin.reflect.KProperty

operator fun RecipeInput.get(index: Int): ItemStack = getItem(index)
operator fun CraftingInput.get(x: Int, y: Int): ItemStack = getItem(x, y)

operator fun CraftingInput.Positioned.get(x: Int, y: Int): ItemStack {
	val actualX = x - left
	val actualY = y - top

	if (actualX !in 0 until input.width() || actualY !in 0 until input.height())
		return ItemStack.EMPTY

	return input.getItem(actualX, actualY)
}

fun FriendlyByteBuf.writeBigInteger(value: BigInteger) {
	writeByteArray(value.toByteArray())
}

fun FriendlyByteBuf.readBigInteger(byteLimit: Int = 128) = BigInteger(readByteArray(byteLimit))

fun <V> Map<*, V>.asLambdaSupplierArray(): Array<() -> V> {
	val result = arrayOfNulls<() -> V>(size)
	var i = 0

	for (k in keys) {
		result[i++] = { this[k] ?: throw ConcurrentModificationException("Key $k is no longer present in map $this") }
	}

	return result as Array<() -> V>
}

fun <V> Map<*, V>.asSupplierArray(): Array<Supplier<V>> {
	val result = arrayOfNulls<Supplier<V>>(size)
	var i = 0

	for (k in keys) {
		result[i++] = { this[k] ?: throw ConcurrentModificationException("Key $k is no longer present in map $this") }
	}

	return result as Array<Supplier<V>>
}

fun <V> Future<V>.asSupplier(): Supplier<V> {
	return Supplier { get() }
}

fun <V> runInBackground(block: Callable<V>): Future<V> {
	return Util.backgroundExecutor().submit(block)
}

operator fun IItemHandler.get(index: Int): ItemStack = getStackInSlot(index)

operator fun JsonObject.set(s: String, value: JsonElement) = add(s, value)
operator fun JsonObject.set(s: String, value: String) = add(s, JsonPrimitive(value))
operator fun JsonObject.set(s: String, value: Number) = add(s, JsonPrimitive(value))
operator fun JsonObject.set(s: String, value: Boolean) = add(s, JsonPrimitive(value))

inline val FluidStack.isNotEmpty get() = !isEmpty
inline val ItemStack.isNotEmpty get() = !isEmpty

inline var Entity.position: Vec3
	get() = position()
	set(value) { setPos(value) }

inline fun itemAttributes(builder: ItemAttributeModifiers.Builder.() -> Unit): ItemAttributeModifiers {
	val value = ItemAttributeModifiers.builder()
	builder(value)
	return value.build()
}

operator fun <T : Comparable<T>> StateHolder<*, *>.get(property: Property<T>): T {
	return getValue(property)
}

operator fun StateHolder<*, *>.get(property: BlockRotationFreedom): BlockRotation {
	return getValue(property.property)
}

operator fun <S : StateHolder<*, *>, T : Comparable<T>> S.set(property: Property<T>, value: T): S {
	return setValue(property, value) as S
}

fun <T> List<T>.toImmutableList(): List<T> {
	return ImmutableList.copyOf(this)
}

operator fun <R> (() -> R).getValue(thisRef: Any, property: KProperty<*>): R {
	return invoke()
}

fun BlockState.getExplosionResistance(level: BlockGetter, pos: BlockPos): Float {
	return try {
		getExplosionResistance(level, pos, null)
	} catch (err: NullPointerException) {
		block.explosionResistance
		// Потому что возможно какой-либо мод не ожидает что Explosion == null
		// особенно учитывая что интерфейс IForgeBlock не имеет @ParamsAreNonnullByDefault
		// и аргумент не помечен как @Nullable
		// тем самым имеет тип Explosion! который указывается как Explosion? .. Explosion!!
	} catch (err: IllegalArgumentException) {
		block.explosionResistance
	}
}

fun Level.gracefulBlockBreak(blockPos: BlockPos, block: BlockState = getBlockState(blockPos)) {
	val tile = getBlockEntity(blockPos)
	val state = block.fluidState.createLegacyBlock()
	block.onRemove(this, blockPos, state, false)
	Block.dropResources(block, this, blockPos, tile)
	block.block.destroy(this, blockPos, block)
	setBlock(blockPos, state, Block.UPDATE_ALL)
}

fun <E> MutableCollection<E>.addAll(elements: Iterator<E>) {
	for (item in elements) {
		add(item)
	}
}

fun <E> MutableCollection<E>.addAll(elements: Stream<out E>) {
	for (item in elements) {
		add(item)
	}
}

fun <E> Iterable<E>.stream(): Stream<out E> {
	return StreamSupport.stream(this.spliterator(), false)
}

fun <E> Iterator<E>.stream(): Stream<out E> {
	return StreamSupport.stream(Spliterators.spliteratorUnknownSize(this, 0), false)
}

fun Entity.genericPositions(): Collection<Vector> {
	return listOf(
		position,
		eyePosition,
		boundingBox.center
	)
}

fun ResourceLocation.toUUID(): UUID {
	return toString().toUUID()
}

fun String.toUUID(): UUID {
	return UUID.nameUUIDFromBytes(toByteArray(Charsets.UTF_8))
}

fun <T> Array<T>.stream(): Stream<T> = Arrays.stream(this)

inline fun <T> MutableList<out Reference<out T>>.forValidRefs(fn: (T) -> Unit) {
	val iterator = listIterator()

	for (value in iterator) {
		val get = value.get()

		if (get == null) {
			iterator.remove()
		} else {
			fn.invoke(get)
		}
	}
}

inline fun <T> MutableList<out Reference<out T>>.forValidRefsBreak(fn: (T) -> Boolean) {
	val iterator = listIterator()

	for (value in iterator) {
		val get = value.get()

		if (get == null) {
			iterator.remove()
		} else {
			if (fn.invoke(get)) {
				break
			}
		}
	}
}

val ComponentContents.key: String
	get() = (this as? TranslatableContents ?: throw ClassCastException("$this is not a TranslatableContents")).key

fun <T : Comparable<T>> BlockState.getValueNullable(prop: Property<T>): T? {
	if (hasProperty(prop)) {
		return getValue(prop)
	}

	return null
}

fun <A, B> lazy2(a: () -> A, b: A.() -> B): Supplier<B> {
	val first = lazy(a)
	return Supplier { b.invoke(first.value) }
}

fun <T : Any> HolderLookup.Provider.lookupOrThrow(key: ResourceKey<T>): Holder<T> {
	return lookupOrThrow(key.registryKey()).getOrThrow(key)
}

fun <T : Any> Registry<T>.getHolder(value: T): Holder<T>? {
	// this is so stupid
	return getHolder(getResourceKey(value).getOrNull() ?: return null).getOrNull()
}

fun <T : Any> Registry<T>.getHolderOrThrow(value: T): Holder<T> {
	// this is so stupid
	return getHolder(getResourceKey(value).orElseThrow()).orElseThrow()
}

// forge registry functionality emulation on vanilla registry
fun <T : Any> Registry<T>.getReverseTag(value: T): Stream<TagKey<T>> {
	return getHolder(value)?.tags() ?: Stream.empty()
}

// forge > neoforge moment
infix fun FluidStack.isSameFluidSameComponents(other: FluidStack): Boolean {
	return FluidStack.isSameFluidSameComponents(this, other)
}

infix fun FluidStack.isSameFluid(other: FluidStack): Boolean {
	return FluidStack.isSameFluid(this, other)
}

infix fun FluidStack.isNotSameFluidSameComponents(other: FluidStack): Boolean {
	return !FluidStack.isSameFluidSameComponents(this, other)
}

infix fun FluidStack.isNotSameFluid(other: FluidStack): Boolean {
	return !FluidStack.isSameFluid(this, other)
}

infix fun FluidStack.isSameAs(other: FluidStack): Boolean {
	return FluidStack.isSameFluidSameComponents(this, other) && amount == other.amount
}

infix fun FluidStack.isNotSameAs(other: FluidStack): Boolean {
	return !FluidStack.isSameFluidSameComponents(this, other) && amount == other.amount
}

fun <K : Any, V> SimpleCache(size: Long, freshness: Duration): Cache<K, V> {
	return Caffeine.newBuilder()
		.maximumSize(size)
		.scheduler(Scheduler.systemScheduler())
		.executor(Util.backgroundExecutor())
		.expireAfterWrite(freshness)
		.build()
}

fun <K : Any, V> SimpleCache(freshness: Duration): Cache<K, V> {
	return SimpleCache(16384L, freshness)
}
